/*
 * Copyright (C) 2005-2014 Andre Noll <maan@systemlinux.org>
 *
 * Licensed under the GPL v2. For licencing details see COPYING.
 */

/** \file audioc.c The client program used to connect to para_audiod. */

#include <netinet/in.h>
#include <sys/socket.h>
#include <regex.h>
#include <sys/types.h>
#include <arpa/inet.h>
#include <sys/un.h>
#include <netdb.h>
#include <stdbool.h>
#include <signal.h>

#include "audioc.cmdline.h"
#include "para.h"
#include "error.h"
#include "net.h"
#include "string.h"
#include "fd.h"
#include "ggo.h"
#include "version.h"

INIT_AUDIOC_ERRLISTS;

/** The gengetopt structure containing command line args. */
static struct audioc_args_info conf;
static char *socket_name;


static int loglevel;
INIT_STDERR_LOGGING(loglevel);

static char *concat_args(unsigned argc, char * const *argv)
{
	int i;
	char *buf = NULL;

	for (i = 0; i < argc; i++) {
		buf = para_strcat(buf, argv[i]);
		if (i != argc - 1)
			buf = para_strcat(buf, "\n");
	}
	return buf;
}

static int connect_audiod(const char *sname, char *args)
{
	int fd = -1, ret;

	ret = connect_local_socket(sname);
	if (ret < 0)
		goto fail;
	fd = ret;
	ret = send_cred_buffer(fd, args);
	if (ret < 0)
		goto fail;
	return fd;
fail:
	PARA_ERROR_LOG("could not connect %s\n", sname);
	if (fd >= 0)
		close(fd);
	return ret;
}

#ifdef HAVE_READLINE
#include "list.h"
#include "sched.h"
#include "buffer_tree.h"
#include "interactive.h"
#include "audiod_completion.h"

static struct sched sched;

struct audioc_task {
	int fd;
	struct btr_node *btrn;
	struct task *task;
};

static struct i9e_completer audiod_completers[];

I9E_DUMMY_COMPLETER(cycle);
I9E_DUMMY_COMPLETER(off);
I9E_DUMMY_COMPLETER(on);
I9E_DUMMY_COMPLETER(sb);
I9E_DUMMY_COMPLETER(tasks);
I9E_DUMMY_COMPLETER(term);

static void help_completer(struct i9e_completion_info *ci,
		struct i9e_completion_result *result)
{
	result->matches = i9e_complete_commands(ci->word, audiod_completers);
}

static void stat_completer(struct i9e_completion_info *ci,
		struct i9e_completion_result *cr)
{
	char *sia[] = {STATUS_ITEM_ARRAY NULL};
	char *opts[] = {"-p", NULL};

	if (ci->word_num <= 2 && ci->word && ci->word[0] == '-')
		i9e_complete_option(opts, ci, cr);
	else
		i9e_extract_completions(ci->word, sia, &cr->matches);
}

static void grab_completer(struct i9e_completion_info *ci,
		struct i9e_completion_result *cr)
{
	char *opts[] = {"-ms", "-ms", "-ma", "-p=", "-n=", "-o", NULL};
	i9e_complete_option(opts, ci, cr);
}

static struct i9e_completer audiod_completers[] = {
	AUDIOD_COMPLETERS
	{.name = NULL}
};

static void audioc_pre_select(struct sched *s, struct task *t)
{
	struct audioc_task *at = task_context(t);
	int ret = btr_node_status(at->btrn, 0, BTR_NT_ROOT);

	if (ret < 0)
		sched_min_delay(s);
	para_fd_set(at->fd, &s->rfds, &s->max_fileno);
}

static int audioc_post_select(struct sched *s, struct task *t)
{
	char *buf = NULL;
	struct audioc_task *at = task_context(t);
	int ret = btr_node_status(at->btrn, 0, BTR_NT_ROOT);

	if (ret < 0)
		goto out;
	if (!FD_ISSET(at->fd, &s->rfds))
		return 0;
	buf = para_malloc(conf.bufsize_arg);
	ret = recv_bin_buffer(at->fd, buf, conf.bufsize_arg);
	PARA_DEBUG_LOG("recv: %d\n", ret);
	if (ret == 0)
		ret = -E_AUDIOC_EOF;
	if (ret < 0)
		goto out;
	btr_add_output(buf, ret, at->btrn);
	return 0;
out:
	if (ret < 0) {
		free(buf);
		btr_remove_node(&at->btrn);
		close(at->fd);
	}
	return ret;
}

static struct audioc_task audioc_task, *at = &audioc_task;

static int audioc_i9e_line_handler(char *line)
{
	char *args = NULL;
	int ret;

	PARA_DEBUG_LOG("line: %s\n", line);
	ret = create_argv(line, " ", &conf.inputs);
	if (ret < 0)
		return ret;
	conf.inputs_num = ret;
	args = concat_args(conf.inputs_num, conf.inputs);
	free_argv(conf.inputs);
	if (!args)
		return 0;
	conf.inputs_num = 0; /* required for audioc_cmdline_parser_free() */
	ret = connect_audiod(socket_name, args);
	if (ret < 0)
		goto out;
	at->fd = ret;
	ret = mark_fd_nonblocking(at->fd);
	if (ret < 0)
		goto close;
	free(args);
	args = NULL;
	at->btrn = btr_new_node(&(struct btr_node_description)
		EMBRACE(.name = "audioc line handler"));
	at->task = task_register(&(struct task_info) {
		.name = "audioc",
		.pre_select = audioc_pre_select,
		.post_select = audioc_post_select,
		.context = at,
	}, &sched);
	i9e_attach_to_stdout(at->btrn);
	return 1;
close:
	close(at->fd);
out:
	free(args);
	return ret;
}

__noreturn static void interactive_session(void)
{
	int ret;
	char *history_file;
	struct sigaction act;
	struct i9e_client_info ici = {
		.fds = {0, 1, 2},
		.prompt = "para_audioc> ",
		.line_handler = audioc_i9e_line_handler,
		.loglevel = loglevel,
		.completers = audiod_completers,
	};
	PARA_NOTICE_LOG("\n%s\n", version_text("audioc"));
	if (conf.history_file_given)
		history_file = para_strdup(conf.history_file_arg);
	else {
		char *home = para_homedir();
		history_file = make_message("%s/.paraslash/audioc.history",
			home);
		free(home);
	}
	ici.history_file = history_file;

	act.sa_handler = i9e_signal_dispatch;
	sigemptyset(&act.sa_mask);
	act.sa_flags = 0;
	sigaction(SIGINT, &act, NULL);
	sched.select_function = i9e_select;

	sched.default_timeout.tv_sec = 1;
	ret = i9e_open(&ici, &sched);
	if (ret < 0)
		goto out;
	para_log = i9e_log;
	ret = schedule(&sched);
	sched_shutdown(&sched);
	i9e_close();
	para_log = stderr_log;
out:
	free(history_file);
	audioc_cmdline_parser_free(&conf);
	if (ret < 0)
		PARA_ERROR_LOG("%s\n", para_strerror(-ret));
	exit(ret < 0? EXIT_FAILURE : EXIT_SUCCESS);
}

__noreturn static void print_completions(void)
{
	int ret = i9e_print_completions(audiod_completers);
	exit(ret <= 0? EXIT_FAILURE : EXIT_SUCCESS);
}

#else /* HAVE_READLINE */

__noreturn static void interactive_session(void)
{
	PARA_EMERG_LOG("interactive sessions not available\n");
	exit(EXIT_FAILURE);
}

__noreturn static void print_completions(void)
{
	PARA_EMERG_LOG("command completion not available\n");
	exit(EXIT_FAILURE);
}

#endif /* HAVE_READLINE */

static char *configfile_exists(void)
{
	char *config_file;
	struct stat statbuf;
	char *home = para_homedir();

	config_file = make_message("%s/.paraslash/audioc.conf", home);
	free(home);
	if (!stat(config_file, &statbuf))
		return config_file;
	free(config_file);
	return NULL;
}

__noreturn static void print_help_and_die(void)
{
	struct ggo_help h = DEFINE_GGO_HELP(audioc);
	bool d = conf.detailed_help_given;

	ggo_print_help(&h, d? GPH_STANDARD_FLAGS_DETAILED : GPH_STANDARD_FLAGS);
	exit(0);
}

/**
 * The client program to connect to para_audiod.
 *
 * \param argc Usual argument count.
 * \param argv Usual argument vector.
 *
 * It connects to the "well-known" local socket to communicate with
 * para_audiod. Authentication is performed by sending a ucred buffer
 * containing the user id to the local socket.
 *
 * Any data received from the socket is written to stdout.
 *
 * \return EXIT_SUCCESS or EXIT_FAILURE.
 *
 * \sa send_cred_buffer(), para_audioc(1), para_audiod(1).
 */
int main(int argc, char *argv[])
{
	int ret, fd;
	char *cf, *buf = NULL, *args = NULL;
	size_t bufsize;

	audioc_cmdline_parser(argc, argv, &conf);
	loglevel = get_loglevel_by_name(conf.loglevel_arg);
	version_handle_flag("audioc", conf.version_given);
	if (conf.help_given || conf.detailed_help_given)
		print_help_and_die();
	cf = configfile_exists();
	if (cf) {
		struct audioc_cmdline_parser_params params = {
			.override = 0,
			.initialize = 0,
			.check_required = 0,
			.check_ambiguity = 0,
			.print_errors = 1,

		};
		audioc_cmdline_parser_config_file(cf, &conf, &params);
		free(cf);
		loglevel = get_loglevel_by_name(conf.loglevel_arg);
	}
	if (conf.socket_given)
		socket_name = para_strdup(conf.socket_arg);
	else {
		char *hn = para_hostname();
		socket_name = make_message("/var/paraslash/audiod_socket.%s",
			hn);
		free(hn);
	}

	if (conf.complete_given)
		print_completions();

	if (conf.inputs_num == 0)
		interactive_session();
	args = concat_args(conf.inputs_num, conf.inputs);

	ret = connect_audiod(socket_name, args);
	free(socket_name);
	if (ret < 0)
		goto out;
	fd = ret;
	ret = mark_fd_blocking(STDOUT_FILENO);
	if (ret < 0)
		goto out;
	bufsize = conf.bufsize_arg;
	buf = para_malloc(bufsize);
	do {
		size_t n = ret = recv_bin_buffer(fd, buf, bufsize);
		if (ret <= 0)
			break;
		ret = write_all(STDOUT_FILENO, buf, n);
	} while (ret >= 0);
out:
	free(buf);
	free(args);
	if (ret < 0)
		PARA_ERROR_LOG("%s\n", para_strerror(-ret));
	return ret < 0? EXIT_FAILURE : EXIT_SUCCESS;
}
