/*
 * Copyright (C) 2006-2014 Andre Noll <maan@systemlinux.org>
 *
 * Licensed under the GPL v2. For licencing details see COPYING.
 */

/** \file sched.c Paraslash's scheduling functions. */

#include <regex.h>
#include <assert.h>

#include "para.h"
#include "ipc.h"
#include "fd.h"
#include "list.h"
#include "sched.h"
#include "string.h"
#include "time.h"
#include "error.h"

static struct timeval now_struct;
struct timeval *now = &now_struct;

static inline bool timeout_is_zero(struct sched *s)
{
	struct timeval *tv = &s->select_timeout;
	return tv->tv_sec == 0 && tv->tv_usec == 0;
}

static void sched_preselect(struct sched *s)
{
	struct task *t, *tmp;

	list_for_each_entry_safe(t, tmp, &s->task_list, node) {
		if (t->error < 0)
			continue;
		if (t->notification != 0)
			sched_min_delay(s);
		if (t->pre_select)
			t->pre_select(s, t);
	}
}

static void unlink_and_free_task(struct task *t)
{
	PARA_INFO_LOG("freeing task %s\n", t->status);
	list_del(&t->node);
	if (t->owned_by_sched)
		free(t);
}

//#define SCHED_DEBUG 1
static inline void call_post_select(struct sched *s, struct task *t)
{
#ifndef SCHED_DEBUG
	t->error = t->post_select(s, t);
#else
	struct timeval t1, t2, diff;
	unsigned long pst;

	clock_get_realtime(&t1);
	t->error = t->post_select(s, t);
	clock_get_realtime(&t2);
	tv_diff(&t1, &t2, &diff);
	pst = tv2ms(&diff);
	if (pst > 50)
		PARA_WARNING_LOG("%s: post_select time: %lums\n",
			t->status, pst);
#endif
}

static unsigned sched_post_select(struct sched *s)
{
	struct task *t, *tmp;
	unsigned num_running_tasks = 0;

	list_for_each_entry_safe(t, tmp, &s->task_list, node) {
		if (t->error < 0) {
			if (t->dead) /* task has been reaped */
				unlink_and_free_task(t);
			continue;
		}
		call_post_select(s, t);
		t->notification = 0;
		if (t->error < 0) {
			if (!t->owned_by_sched)
				list_del(&t->node);
		} else
			num_running_tasks++;
	}
	return num_running_tasks;
}

/**
 * The core function of all paraslash programs.
 *
 * \param s Pointer to the scheduler struct.
 *
 * This function updates the global \a now pointer, calls all registered
 * pre_select hooks which may set the timeout and add any file descriptors to
 * the fd sets of \a s.  Next, it calls para_select() and makes the result available
 * to the registered tasks by calling their post_select hook.
 *
 * \return Zero if no more tasks are left in the task list, negative if the
 * select function returned an error.
 *
 * \sa \ref task, \ref now.
 */
int schedule(struct sched *s)
{
	int ret;
	unsigned num_running_tasks;

	if (!s->select_function)
		s->select_function = para_select;
again:
	FD_ZERO(&s->rfds);
	FD_ZERO(&s->wfds);
	s->select_timeout = s->default_timeout;
	s->max_fileno = -1;
	clock_get_realtime(now);
	sched_preselect(s);
	ret = s->select_function(s->max_fileno + 1, &s->rfds, &s->wfds,
		&s->select_timeout);
	if (ret < 0)
		return ret;
	if (ret == 0) {
		/*
		 * APUE: Be careful not to check the descriptor sets on return
		 * unless the return value is greater than zero. The return
		 * state of the descriptor sets is implementation dependent if
		 * either a signal is caught or the timer expires.
		 */
		FD_ZERO(&s->rfds);
		FD_ZERO(&s->wfds);
	}
	clock_get_realtime(now);
	num_running_tasks = sched_post_select(s);
	if (num_running_tasks == 0)
		return 0;
	goto again;
}

/**
 * Obtain the error status of a task and deallocate its resources.
 *
 * \param tptr Identifies the task to reap.
 *
 * This function is similar to wait(2) in that it returns information about a
 * terminated task and allows to release the resources associated with the
 * task. Until this function is called, the terminated task remains in a zombie
 * state.
 *
 * \return If \a tptr is \p NULL, or \a *tptr is \p NULL, the function does
 * nothing and returns zero. Otherwise, it is checked whether the task
 * identified by \a tptr is still running. If it is, the function returns zero
 * and again, no action is taken. Otherwise the (negative) error code of the
 * terminated task is returned and \a *tptr is set to \p NULL. The task will
 * then be removed removed from the scheduler task list.
 *
 * \sa \ref sched_shutdown(), wait(2).
 */
int task_reap(struct task **tptr)
{
	struct task *t;

	if (!tptr)
		return 0;
	t = *tptr;
	if (!t)
		return 0;
	if (!t->owned_by_sched)
		return 0;
	if (t->error >= 0)
		return 0;
	if (t->dead) /* will be freed in sched_post_select() */
		return 0;
	/*
	 * With list_for_each_entry_safe() it is only safe to remove the
	 * _current_ list item. Since we are being called from the loop in
	 * schedule() via some task's ->post_select() function, freeing the
	 * given task here would result in use-after-free bugs in schedule().
	 * So we only set t->dead which tells schedule() to free the task in
	 * the next iteration of its loop.
	 */
	t->dead = true;
	*tptr = NULL;
	return t->error;
}

/**
 * Deallocate all resources of all tasks of a scheduler instance.
 *
 * \param s The scheduler instance.
 *
 * This should only be called after \ref schedule() has returned.
 */
void sched_shutdown(struct sched *s)
{
	struct task *t, *tmp;

	list_for_each_entry_safe(t, tmp, &s->task_list, node) {
		if (t->error >= 0)
			/* The task list should contain only terminated tasks. */
			PARA_WARNING_LOG("shutting down running task %s\n",
				t->status);
		unlink_and_free_task(t);
	}
}

/**
 * Add a task to the scheduler. Deprecated.
 *
 * \param t The task to add.
 * \param s The scheduler instance to add the task to.
 *
 * \sa task::pre_select, task::post_select
 */
void register_task(struct sched *s, struct task *t)
{
	PARA_INFO_LOG("registering %s (%p)\n", t->status, t);
	assert(t->post_select);
	t->notification = 0;
	t->owned_by_sched = false;
	if (!s->task_list.next)
		INIT_LIST_HEAD(&s->task_list);
	list_add_tail(&t->node, &s->task_list);
}

/**
 * Add a task to the scheduler task list.
 *
 * \param info Task information supplied by the caller.
 * \param s The scheduler instance.
 *
 * \return A pointer to a newly allocated task structure. It will be
 * freed by sched_shutdown().
 */
struct task *task_register(struct task_info *info, struct sched *s)
{
	struct task *t = para_malloc(sizeof(*t));

	assert(info->post_select);

	if (!s->task_list.next)
		INIT_LIST_HEAD(&s->task_list);

	snprintf(t->status, sizeof(t->status) - 1, "%s", info->name);
	t->status[sizeof(t->status) - 1] = '\0';
	t->notification = 0;
	t->error = 0;
	t->dead = false;
	t->pre_select = info->pre_select;
	t->post_select = info->post_select;
	t->context = info->context;
	t->owned_by_sched = true;
	list_add_tail(&t->node, &s->task_list);
	return t;
}

/**
 * Obtain the context pointer of a task.
 *
 * \param t Return this task's context pointer.
 *
 * \return A pointer to the memory location specified previously as \a
 * task_info->context when the task was registered with \ref task_register().
 */
void *task_context(struct task *t)
{
	assert(t->owned_by_sched);
	return t->context;
}

/**
 * Get the list of all registered tasks.
 *
 * \param s The scheduler instance to get the task list from.
 *
 * \return The task list.
 *
 * Each entry of the list contains an identifier which is simply a hex number.
 * The result is dynamically allocated and must be freed by the caller.
 */
char *get_task_list(struct sched *s)
{
	struct task *t, *tmp;
	char *msg = NULL;

	list_for_each_entry_safe(t, tmp, &s->task_list, node) {
		char *tmp_msg;
		tmp_msg = make_message("%s%p\t%s\t%s\n", msg? msg : "", t,
			t->error < 0? (t->dead? "dead" : "zombie") : "running",
			t->status);
		free(msg);
		msg = tmp_msg;
	}
	return msg;
}

/**
 * Set the notification value of a task.
 *
 * \param t The task to notify.
 * \param err A positive error code.
 *
 * Tasks which honor notifications are supposed to call \ref
 * task_get_notification() in their post_select function and act on the
 * returned notification value.
 *
 * If the scheduler detects during its pre_select loop that at least one task
 * has been notified, the loop terminates, and the post_select methods of all
 * taks are immediately called again.
 *
 * The notification for a task is reset after the call to its post_select
 * method.
 *
 * \sa \ref task_get_notification().
 */
void task_notify(struct task *t, int err)
{
	assert(err > 0);
	if (t->notification == -err) /* ignore subsequent notifications */
		return;
	PARA_INFO_LOG("notifying task %s: %s\n", t->status, para_strerror(err));
	t->notification = -err;
}

/**
 * Return the notification value of a task.
 *
 * \param t The task to get the notification value from.
 *
 * \return The notification value. If this is negative, the task has been
 * notified by another task. Tasks are supposed to check for notifications by
 * calling this function from their post_select method.
 *
 * \sa \ref task_notify().
 */
int task_get_notification(const struct task *t)
{
	return t->notification;
}

/**
 * Set the notification value of all tasks of a scheduler instance.
 *
 * \param s The scheduler instance whose tasks should be notified.
 * \param err A positive error code.
 *
 * This simply iterates over all existing tasks of \a s and sets each
 * task's notification value to \p -err.
 */
void task_notify_all(struct sched *s, int err)
{
	struct task *t;

	list_for_each_entry(t, &s->task_list, node)
		task_notify(t, err);
}

/**
 * Set the select timeout to the minimal possible value.
 *
 * \param s Pointer to the scheduler struct.
 *
 * This causes the next select() call to return immediately.
 */
void sched_min_delay(struct sched *s)
{
	s->select_timeout.tv_sec = s->select_timeout.tv_usec = 0;
}

/**
 * Impose an upper bound for the timeout of the next select() call.
 *
 * \param to Maximal allowed timeout.
 * \param s Pointer to the scheduler struct.
 *
 * If the current scheduler timeout is already smaller than \a to, this
 * function does nothing. Otherwise the timeout for the next select() call is
 * set to the given value.
 *
 * \sa sched_request_timeout_ms().
 */
void sched_request_timeout(struct timeval *to, struct sched *s)
{
	if (tv_diff(&s->select_timeout, to, NULL) > 0)
		s->select_timeout = *to;
}

/**
 * Force the next select() call to return before the given amount of milliseconds.
 *
 * \param ms The maximal allowed timeout in milliseconds.
 * \param s Pointer to the scheduler struct.
 *
 * Like sched_request_timeout() this imposes an upper bound on the timeout
 * value for the next select() call.
 */
void sched_request_timeout_ms(long unsigned ms, struct sched *s)
{
	struct timeval tv;
	ms2tv(ms, &tv);
	sched_request_timeout(&tv, s);
}

/**
 * Force the next select() call to return before the given future time.
 *
 * \param barrier Absolute time before select() should return.
 * \param s Pointer to the scheduler struct.
 *
 * \return If \a barrier is in the past, this function does nothing and returns
 * zero. Otherwise it returns one.
 *
 * \sa sched_request_barrier_or_min_delay().
 */
int sched_request_barrier(struct timeval *barrier, struct sched *s)
{
	struct timeval diff;

	if (tv_diff(now, barrier, &diff) > 0)
		return 0;
	sched_request_timeout(&diff, s);
	return 1;
}

/**
 * Force the next select() call to return before the given time.
 *
 * \param barrier Absolute time before select() should return.
 * \param s Pointer to the scheduler struct.
 *
 * \return If \a barrier is in the past, this function requests a minimal
 * timeout and returns zero. Otherwise it returns one.
 *
 * \sa sched_min_delay(), sched_request_barrier().
 */
int sched_request_barrier_or_min_delay(struct timeval *barrier, struct sched *s)
{
	struct timeval diff;

	if (tv_diff(now, barrier, &diff) > 0) {
		sched_min_delay(s);
		return 0;
	}
	sched_request_timeout(&diff, s);
	return 1;
}
