/*
 * Copyright (C) 2006-2014 Andre Noll <maan@systemlinux.org>
 *
 * Licensed under the GPL v2. For licencing details see COPYING.
 */

/** \file sched.h Sched and task structures and exported symbols from sched.c. */


/**
 * Paraslash's scheduler.
 *
 * Designed with KISS in mind. It maintains a list of task structures which is
 * extended when a new task is registered. Each task may define a pre_select
 * function which is called from the scheduler main loop before it calls
 * select(). Similarly, each task must define a post_select function which is
 * called after the select call.
 */
struct sched {
	/** Initial value before any pre_select call. */
	struct timeval default_timeout;
	/** The current timeout for the upcoming select call. */
	struct timeval select_timeout;
	/** fds that should be watched for readability. */
	fd_set rfds;
	/** fds that should be watched for writability. */
	fd_set wfds;
	/** Highest numbered file descriptor in any of the above fd sets. */
	int max_fileno;
	/** If non-NULL, use this function instead of para_select. */
	int (*select_function)(int, fd_set *, fd_set *, struct timeval *);
	/** Tasks which have been registered to the scheduler. */
	struct list_head task_list;
};

/**
 * Paraslash's task structure.
 *
 * This is considered an internal structure and will eventually be made private.
 *
 * \sa \ref sched.
 */
struct task {
	/** Copied from the task_info struct during task_register(). */
	void (*pre_select)(struct sched *s, struct task *t);
	/** Copied from the task_info struct during task_register(). */
	int (*post_select)(struct sched *s, struct task *t);
	/** Whether this task is active (>=0) or in error state (<0). */
	int error;
	/** Position of the task in the task list of the scheduler. */
	struct list_head node;
	/** The task name supplied when the task was registered(). */
	char status[255];
	/** If less than zero, the task was notified by another task. */
	int notification;
	/** Whether the task structure should be freed in sched_shutdown(). */
	bool owned_by_sched;
	/** True if task is in error state and exit status has been queried. */
	bool dead;
	/** Usually a pointer to the struct containing this task. */
	void *context;
};

/** Information that must be supplied by callers of \ref task_register(). */
struct task_info {
	/** Used for log messages and by \ref get_task_list(). */
	const char *name;
	/**
	 * The optional pre select method.
	 *
	 * Its purpose is to add file descriptors to the fd sets of the
	 * scheduler and to decrease the select timeout if necessary.
	 */
	void (*pre_select)(struct sched *s, struct task *t);
	/**
	 * The mandatory post select method.
	 *
	 * Its purpose is to evaluate and act upon the results of the previous
	 * select call. If this function returns a negative value, the
	 * scheduler unregisters the task.
	 */
	int (*post_select)(struct sched *s, struct task *t);
	/**
	 * This pointer is saved when the task is register(ed). It may be
	 * queried from ->pre_select() and ->post_select() via \ref
	 * task_context(). Usually this is a pointer to the struct owned by the
	 * caller which contains the task pointer as one member.
	 */
	void *context;
};

/**
 * This is set by the scheduler at the beginning of its main loop.  It may be
 * used (read-only) from everywhere. As none of the functions called by the
 * scheduler are allowed to block, this value should be accurate enough so that
 * there is no need to call clock_gettime() directly.
 */
extern struct timeval *now;

struct task *task_register(struct task_info *info, struct sched *s);
void *task_context(struct task *t);
void register_task(struct sched *s, struct task *t);
int schedule(struct sched *s);
void sched_shutdown(struct sched *s);
char *get_task_list(struct sched *s);
void task_notify(struct task *t, int err);
void task_notify_all(struct sched *s, int err);
int task_get_notification(const struct task *t);
int task_reap(struct task **tptr);
void sched_min_delay(struct sched *s);
void sched_request_timeout(struct timeval *to, struct sched *s);
void sched_request_timeout_ms(long unsigned ms, struct sched *s);
int sched_request_barrier(struct timeval *barrier, struct sched *s);
int sched_request_barrier_or_min_delay(struct timeval *barrier, struct sched *s);
